#  Copyright (c) 2024, Salesforce, Inc.
#  SPDX-License-Identifier: Apache-2.0

from collections.abc import Callable
from functools import partial
from typing import Optional

import torch
import torch.nn.functional as F
from torch import nn

from .attention import GroupedQueryAttention
from .ffn import FeedForward, GatedLinearUnitFeedForward, MoEFeedForward
from .position import AttentionBias, QueryKeyProjection


class TransformerEncoderLayer(nn.Module):
    def __init__(
        self,
        self_attn: GroupedQueryAttention,
        ffn: FeedForward,
        norm1: Optional[nn.Module],
        norm2: Optional[nn.Module],
        post_attn_dropout_p: float = 0.0,
        pre_norm: bool = True,
    ):
        super().__init__()
        self.pre_norm = pre_norm
        self.dropout_p = post_attn_dropout_p

        self.self_attn = self_attn
        self.ffn = ffn
        self.norm1 = norm1 or nn.Identity()
        self.norm2 = norm2 or nn.Identity()
        self.dropout = nn.Dropout(post_attn_dropout_p)

    def forward(self, x, attn_mask=None, var_id=None, time_id=None, centroid=None):
        if self.pre_norm:
            x = x + self._sa_block(
                self.norm1(x), attn_mask, var_id=var_id, time_id=time_id
            )
            x = x + self.ffn(self.norm2(x), centroid=centroid)
        else:
            x = self.norm1(
                x + self._sa_block(x, attn_mask, var_id=var_id, time_id=time_id)
            )
            x = self.norm2(x + self.ffn(x, centroid=centroid))

        return x

    def _sa_block(self, x, attn_mask, var_id=None, time_id=None):
        x = self.self_attn(
            x, x, x,
            attn_mask=attn_mask,
            query_var_id=var_id,
            kv_var_id=var_id,
            query_time_id=time_id,
            kv_time_id=time_id,
        )
        return self.dropout(x)


class TransformerEncoder(nn.Module):
    def __init__(
        self,
        d_model: int,
        num_layers: int,
        num_heads: Optional[int] = None,
        num_groups: Optional[int] = None,
        pre_norm: bool = True,
        attn_dropout_p: float = 0.0,
        dropout_p: float = 0.0,
        norm_layer: Optional[Callable[[int], nn.Module]] = nn.LayerNorm,
        activation: Callable[[torch.Tensor], torch.Tensor] = F.silu,
        use_moe: bool = False,
        use_glu: bool = True,
        use_qk_norm: bool = True,
        var_attn_bias_layer: Optional[Callable] = None,
        time_attn_bias_layer: Optional[Callable] = None,
        var_qk_proj_layer: Optional[Callable] = None,
        time_qk_proj_layer: Optional[Callable] = None,
        shared_var_attn_bias: bool = False,
        shared_time_attn_bias: bool = False,
        shared_var_qk_proj: bool = False,
        shared_time_qk_proj: bool = False,
        d_ff: Optional[int] = None,
    ):
        super().__init__()
        self.use_moe = use_moe
        num_heads = num_heads or d_model // 64
        num_groups = num_groups or num_heads

        var_attn_bias = self.get_layer(
            d_model, num_heads, num_groups, var_attn_bias_layer, shared_var_attn_bias,
        )
        time_attn_bias = self.get_layer(
            d_model, num_heads, num_groups, time_attn_bias_layer, shared_time_attn_bias,
        )
        var_qk_proj = self.get_layer(
            d_model, num_heads, num_groups, var_qk_proj_layer, shared_var_qk_proj,
        )
        time_qk_proj = self.get_layer(
            d_model, num_heads, num_groups, time_qk_proj_layer, shared_time_qk_proj,
        )

        get_self_attn = partial(
            GroupedQueryAttention,
            dim=d_model,
            num_heads=num_heads,
            num_groups=num_groups,
            bias=False,
            norm_layer=norm_layer if use_qk_norm else None,
            softmax_scale=None,
            attn_dropout_p=attn_dropout_p,
            var_attn_bias=var_attn_bias,
            time_attn_bias=time_attn_bias,
            var_qk_proj=var_qk_proj,
            time_qk_proj=time_qk_proj,
        )
        if not use_moe:
            get_ffn = partial(
                GatedLinearUnitFeedForward if use_glu else FeedForward,
                in_dim=d_model,
                hidden_dim=d_ff,
                out_dim=None,
                activation=activation,
                bias=False,
                ffn_dropout_p=dropout_p,
            )
        else:
            get_ffn = partial(
                MoEFeedForward,
                num_experts=32,
                num_experts_per_token=2,
                in_dim=d_model,
                hidden_dim=d_ff,
                out_dim=None,
                activation=activation,
                bias=False,
                ffn_dropout_p=dropout_p,
            )
            self.register_buffer(
                "centroid", torch.empty(num_layers, 32, d_model, dtype=torch.float64)
            )
        get_encoder_layer_norm = partial(norm_layer, d_model)

        self.layers = nn.ModuleList(
            [
                TransformerEncoderLayer(
                    self_attn=get_self_attn(),
                    ffn=get_ffn(),
                    norm1=get_encoder_layer_norm(),
                    norm2=get_encoder_layer_norm(),
                    pre_norm=pre_norm,
                    post_attn_dropout_p=dropout_p,
                )
                for _ in range(num_layers)
            ]
        )
        self.norm = norm_layer(d_model)

    @staticmethod
    def get_layer(dim, num_heads, num_groups, layer, shared_layer):
        if layer is None:
            return None
        if shared_layer:
            module = layer(dim=dim, num_heads=num_heads, num_groups=num_groups)
            return lambda: module
        return partial(layer, dim=dim, num_heads=num_heads, num_groups=num_groups)

    def forward(self, x, attn_mask=None, var_id=None, time_id=None):
        if self.use_moe:
            for idx, layer in enumerate(self.layers):
                x = layer(x, attn_mask, var_id=var_id, time_id=time_id, centroid=self.centroid[idx])
        else:
            for layer in self.layers:
                x = layer(x, attn_mask, var_id=var_id, time_id=time_id)
        return self.norm(x)



"""
Tests for MoiraiForecaster
"""

import numpy as np
import pandas as pd
import pytest
from moirai_sklearn import MoiraiForecaster


@pytest.fixture
def simple_timeseries():
    """Create a simple sine wave time series for testing."""
    return np.sin(np.linspace(0, 10, 100)) + np.random.RandomState(42).randn(100) * 0.1


@pytest.fixture
def forecaster():
    """Create a forecaster instance."""
    return MoiraiForecaster(model_name="Salesforce/moirai-2.0-R-small")


class TestMoiraiForecaster:
    
    def test_init(self):
        """Test forecaster initialization."""
        model = MoiraiForecaster()
        assert model.model_name == "Salesforce/moirai-2.0-R-small"
        assert model.context_length == 1000
        assert model.device == "auto"
    
    def test_predict_with_numpy(self, forecaster, simple_timeseries):
        """Test predict with numpy array input."""
        predictions = forecaster.predict(simple_timeseries, horizon=10)
        assert predictions.shape == (10,)
        assert not np.isnan(predictions).any()
    
    def test_predict_with_pandas_series(self, forecaster, simple_timeseries):
        """Test predict with pandas Series input."""
        series = pd.Series(simple_timeseries)
        predictions = forecaster.predict(series, horizon=10)
        assert predictions.shape == (10,)
        assert not np.isnan(predictions).any()
    
    def test_predict_with_dataframe(self, forecaster, simple_timeseries):
        """Test predict with pandas DataFrame input."""
        df = pd.DataFrame({'value': simple_timeseries})
        predictions = forecaster.predict(df, horizon=10)
        assert predictions.shape == (10,)
        assert not np.isnan(predictions).any()
    
    def test_predict_median(self, forecaster, simple_timeseries):
        """Test predict_median."""
        median = forecaster.predict_median(simple_timeseries, horizon=10)
        assert median.shape == (10,)
        assert not np.isnan(median).any()
    
    def test_predict_mean(self, forecaster, simple_timeseries):
        """Test predict_mean."""
        mean = forecaster.predict_mean(simple_timeseries, horizon=10)
        assert mean.shape == (10,)
        assert not np.isnan(mean).any()
    
    def test_predict_mode(self, forecaster, simple_timeseries):
        """Test predict_mode."""
        mode = forecaster.predict_mode(simple_timeseries, horizon=10)
        assert mode.shape == (10,)
        assert not np.isnan(mode).any()
    
    def test_predict_quantile_single(self, forecaster, simple_timeseries):
        """Test predict_quantile with single quantile."""
        q50 = forecaster.predict_quantile(simple_timeseries, horizon=10, q=0.5)
        assert q50.shape == (10,)
        assert not np.isnan(q50).any()
    
    def test_predict_quantile_multiple(self, forecaster, simple_timeseries):
        """Test predict_quantile with multiple quantiles."""
        quantiles = forecaster.predict_quantile(simple_timeseries, horizon=10, q=[0.1, 0.5, 0.9])
        assert quantiles.shape == (10, 3)
        assert not np.isnan(quantiles).any()
        # Check ordering: q10 <= q50 <= q90
        assert (quantiles[:, 0] <= quantiles[:, 1]).all()
        assert (quantiles[:, 1] <= quantiles[:, 2]).all()
    
    def test_predict_quantile_invalid(self, forecaster, simple_timeseries):
        """Test predict_quantile with invalid quantile raises error."""
        with pytest.raises(ValueError, match="Quantile.*not available"):
            forecaster.predict_quantile(simple_timeseries, horizon=10, q=0.99)
    
    def test_predict_interval_80(self, forecaster, simple_timeseries):
        """Test predict_interval with 80% confidence."""
        intervals = forecaster.predict_interval(simple_timeseries, horizon=10, confidence=0.8)
        assert intervals.shape == (10, 2)
        assert not np.isnan(intervals).any()
        # Check lower <= upper
        assert (intervals[:, 0] <= intervals[:, 1]).all()
    
    def test_predict_interval_60(self, forecaster, simple_timeseries):
        """Test predict_interval with 60% confidence."""
        intervals = forecaster.predict_interval(simple_timeseries, horizon=10, confidence=0.6)
        assert intervals.shape == (10, 2)
        assert not np.isnan(intervals).any()
        # Check lower <= upper
        assert (intervals[:, 0] <= intervals[:, 1]).all()
    
    def test_predict_interval_invalid(self, forecaster, simple_timeseries):
        """Test predict_interval with invalid confidence raises error."""
        with pytest.raises(ValueError, match="Confidence level.*not supported"):
            forecaster.predict_interval(simple_timeseries, horizon=10, confidence=0.95)
    
    def test_predict_std(self, forecaster, simple_timeseries):
        """Test predict_std."""
        std = forecaster.predict_std(simple_timeseries, horizon=10)
        assert std.shape == (10,)
        assert not np.isnan(std).any()
        assert (std > 0).all()  # Standard deviation should be positive
    
    def test_predict_all(self, forecaster, simple_timeseries):
        """Test predict_all returns comprehensive DataFrame."""
        df = forecaster.predict_all(simple_timeseries, horizon=10)
        assert isinstance(df, pd.DataFrame)
        assert len(df) == 10
        
        # Check expected columns
        expected_cols = ['step', 'mean', 'median', 'std', 
                        'p10', 'p20', 'p30', 'p40', 'p50', 
                        'p60', 'p70', 'p80', 'p90']
        assert all(col in df.columns for col in expected_cols)
        
        # Check no NaN values
        assert not df.isnull().any().any()
        
        # Check quantile ordering
        assert (df['p10'] <= df['p50']).all()
        assert (df['p50'] <= df['p90']).all()
    
    def test_different_horizons(self, forecaster, simple_timeseries):
        """Test prediction with different horizons."""
        for horizon in [5, 10, 20, 50]:
            predictions = forecaster.predict(simple_timeseries, horizon=horizon)
            assert predictions.shape == (horizon,)
            assert not np.isnan(predictions).any()
    
    def test_model_caching(self, simple_timeseries):
        """Test that model is cached and reused."""
        forecaster = MoiraiForecaster()
        
        # First prediction - loads model
        _ = forecaster.predict(simple_timeseries, horizon=10)
        assert forecaster._model is not None
        assert forecaster._module is not None
        
        # Second prediction - reuses model
        model_id = id(forecaster._module)
        _ = forecaster.predict(simple_timeseries, horizon=10)
        assert id(forecaster._module) == model_id  # Same module instance
    
    def test_invalid_input_type(self, forecaster):
        """Test that invalid input type raises error."""
        with pytest.raises(ValueError, match="Unsupported data type"):
            forecaster.predict([1, 2, 3, 4, 5], horizon=10)  # List not supported


if __name__ == "__main__":
    pytest.main([__file__, "-v"])



#  Copyright (c) 2024, Salesforce, Inc.
#  SPDX-License-Identifier: Apache-2.0
#
#  Vendored and modified: replaced lightning.LightningModule with plain nn.Module.
#  Removed gluonts dependency (create_predictor / get_default_transform not included).

import math
from contextlib import contextmanager
from copy import deepcopy
from typing import Any, Generator, List, Optional

import numpy as np
import torch
from einops import rearrange, reduce, repeat
from torch import nn

from .imputation import CausalMeanImputation
from .moirai2_module import Moirai2Module


class _HParams(dict):
    """Minimal hparams container mimicking lightning's hparams."""
    def __getattr__(self, key):
        try:
            return self[key]
        except KeyError:
            raise AttributeError(key)

    def __setattr__(self, key, value):
        self[key] = value


class Moirai2Forecast(nn.Module):
    def __init__(
        self,
        prediction_length: int,
        target_dim: int,
        feat_dynamic_real_dim: int,
        past_feat_dynamic_real_dim: int,
        context_length: int,
        module_kwargs: Optional[dict[str, Any]] = None,
        module: Optional[Moirai2Module] = None,
    ):
        assert (module is not None) or (
            module_kwargs is not None
        ), "if module is not provided, module_kwargs is required"
        if module_kwargs and "attn_dropout_p" in module_kwargs:
            module_kwargs["attn_dropout_p"] = 0
        if module_kwargs and "dropout_p" in module_kwargs:
            module_kwargs["dropout_p"] = 0

        super().__init__()
        self.hparams = _HParams(
            prediction_length=prediction_length,
            target_dim=target_dim,
            feat_dynamic_real_dim=feat_dynamic_real_dim,
            past_feat_dynamic_real_dim=past_feat_dynamic_real_dim,
            context_length=context_length,
        )
        self.module = Moirai2Module(**module_kwargs) if module is None else module
        self.module.eval()

    @contextmanager
    def hparams_context(
        self,
        prediction_length=None,
        target_dim=None,
        feat_dynamic_real_dim=None,
        past_feat_dynamic_real_dim=None,
        context_length=None,
    ) -> Generator["Moirai2Forecast", None, None]:
        kwargs = {
            "prediction_length": prediction_length,
            "target_dim": target_dim,
            "feat_dynamic_real_dim": feat_dynamic_real_dim,
            "past_feat_dynamic_real_dim": past_feat_dynamic_real_dim,
            "context_length": context_length,
        }
        old_hparams = deepcopy(self.hparams)
        for kw, arg in kwargs.items():
            if arg is not None:
                self.hparams[kw] = arg

        yield self

        for kw in kwargs:
            self.hparams[kw] = old_hparams[kw]

    @property
    def past_length(self) -> int:
        return self.hparams.context_length

    def context_token_length(self, patch_size: int) -> int:
        return math.ceil(self.hparams.context_length / patch_size)

    def prediction_token_length(self, patch_size) -> int:
        return math.ceil(self.hparams.prediction_length / patch_size)

    def forward(
        self,
        past_target,
        past_observed_target,
        past_is_pad,
        feat_dynamic_real=None,
        observed_feat_dynamic_real=None,
        past_feat_dynamic_real=None,
        past_observed_feat_dynamic_real=None,
    ):
        (
            target,
            observed_mask,
            sample_id,
            time_id,
            variate_id,
            prediction_mask,
        ) = self._convert(
            self.module.patch_size,
            past_target,
            past_observed_target,
            past_is_pad,
            feat_dynamic_real=feat_dynamic_real,
            observed_feat_dynamic_real=observed_feat_dynamic_real,
            past_feat_dynamic_real=past_feat_dynamic_real,
            past_observed_feat_dynamic_real=past_observed_feat_dynamic_real,
        )

        per_var_context_token = self.context_token_length(self.module.patch_size)
        total_context_token = self.hparams.target_dim * per_var_context_token
        per_var_predict_token = self.prediction_token_length(self.module.patch_size)
        total_predict_token = self.hparams.target_dim * per_var_predict_token

        pred_index = torch.arange(
            start=per_var_context_token - 1,
            end=total_context_token,
            step=per_var_context_token,
        )
        assign_index = torch.arange(
            start=total_context_token,
            end=total_context_token + total_predict_token,
            step=per_var_predict_token,
        )
        quantile_prediction = repeat(
            target,
            "... patch_size -> ... num_quantiles patch_size",
            num_quantiles=len(self.module.quantile_levels),
            patch_size=self.module.patch_size,
        ).clone()

        preds = self.module(
            target,
            observed_mask,
            sample_id,
            time_id,
            variate_id,
            prediction_mask,
            training_mode=False,
        )

        def structure_multi_predict(
            per_var_predict_token,
            pred_index,
            assign_index,
            preds,
        ):
            preds = rearrange(
                preds,
                "... (predict_token num_quantiles patch_size) -> ... predict_token num_quantiles patch_size",
                predict_token=self.module.num_predict_token,
                num_quantiles=self.module.num_quantiles,
                patch_size=self.module.patch_size,
            )
            preds = rearrange(
                preds[..., pred_index, :per_var_predict_token, :, :],
                "... pred_index predict_token num_quantiles patch_size -> ... (pred_index predict_token) num_quantiles patch_size",
            )
            adjusted_assign_index = torch.cat(
                [
                    torch.arange(start=idx, end=idx + per_var_predict_token)
                    for idx in assign_index
                ]
            )
            return preds, adjusted_assign_index

        if per_var_predict_token <= self.module.num_predict_token:
            preds, adjusted_assign_index = structure_multi_predict(
                per_var_predict_token,
                pred_index,
                assign_index,
                preds,
            )
            quantile_prediction[..., adjusted_assign_index, :, :] = preds
            return self._format_preds(
                self.module.num_quantiles,
                self.module.patch_size,
                quantile_prediction,
                self.hparams.target_dim,
            )
        else:
            expand_target = repeat(
                target,
                "batch_size ...  -> batch_size num_quantiles ...",
                num_quantiles=len(self.module.quantile_levels),
                batch_size=target.shape[0],
            ).clone()
            expand_prediction_mask = repeat(
                prediction_mask,
                "batch_size ...  -> batch_size num_quantiles ...",
                num_quantiles=len(self.module.quantile_levels),
                batch_size=target.shape[0],
            ).clone()
            expand_observed_mask = repeat(
                observed_mask,
                "batch_size ...  -> batch_size num_quantiles ...",
                num_quantiles=len(self.module.quantile_levels),
                batch_size=target.shape[0],
            ).clone()
            expand_sample_id = repeat(
                sample_id,
                "batch_size ...  -> batch_size num_quantiles ...",
                num_quantiles=len(self.module.quantile_levels),
                batch_size=target.shape[0],
            ).clone()
            expand_time_id = repeat(
                time_id,
                "batch_size ...  -> batch_size num_quantiles ...",
                num_quantiles=len(self.module.quantile_levels),
                batch_size=target.shape[0],
            ).clone()
            expand_variate_id = repeat(
                variate_id,
                "batch_size ...  -> batch_size num_quantiles ...",
                num_quantiles=len(self.module.quantile_levels),
                batch_size=target.shape[0],
            ).clone()

            preds, adjusted_assign_index = structure_multi_predict(
                self.module.num_predict_token,
                pred_index,
                assign_index,
                preds,
            )
            quantile_prediction[..., adjusted_assign_index, :, :] = preds

            expand_target[..., adjusted_assign_index, :] = rearrange(
                preds,
                "... predict_token num_quantiles patch_size -> ... num_quantiles predict_token patch_size",
                num_quantiles=self.module.num_quantiles,
                patch_size=self.module.patch_size,
                predict_token=self.module.num_predict_token,
            )
            expand_prediction_mask[..., adjusted_assign_index] = False

            remain_step = per_var_predict_token - self.module.num_predict_token
            while remain_step > 0:
                preds = self.module(
                    expand_target,
                    expand_observed_mask,
                    expand_sample_id,
                    expand_time_id,
                    expand_variate_id,
                    expand_prediction_mask,
                    training_mode=False,
                )

                pred_index = assign_index + self.module.num_predict_token - 1
                assign_index = pred_index + 1
                preds, adjusted_assign_index = structure_multi_predict(
                    (
                        self.module.num_predict_token
                        if remain_step - self.module.num_predict_token > 0
                        else remain_step
                    ),
                    pred_index,
                    assign_index,
                    preds,
                )
                quantile_prediction_next_step = rearrange(
                    preds,
                    "... num_quantiles_prev pred_index num_quantiles patch_size -> ... pred_index (num_quantiles_prev num_quantiles) patch_size",
                    num_quantiles=self.module.num_quantiles,
                    patch_size=self.module.patch_size,
                )
                quantile_prediction_next_step = torch.quantile(
                    quantile_prediction_next_step,
                    torch.tensor(
                        self.module.quantile_levels,
                        device=self.device,
                        dtype=torch.float32,
                    ),
                    dim=-2,
                )
                quantile_prediction[..., adjusted_assign_index, :, :] = rearrange(
                    quantile_prediction_next_step,
                    "num_quantiles ... patch_size -> ... num_quantiles patch_size",
                )

                expand_target[..., adjusted_assign_index, :] = rearrange(
                    quantile_prediction_next_step,
                    "num_quantiles batch_size predict_token patch_size -> batch_size num_quantiles predict_token patch_size",
                    num_quantiles=self.module.num_quantiles,
                    patch_size=self.module.patch_size,
                    predict_token=len(adjusted_assign_index),
                )
                expand_prediction_mask[..., adjusted_assign_index] = False

                remain_step -= self.module.num_predict_token

            return self._format_preds(
                self.module.num_quantiles,
                self.module.patch_size,
                quantile_prediction,
                self.hparams.target_dim,
            )

    @property
    def device(self):
        """Get device from module parameters (replaces lightning's self.device)."""
        return next(self.parameters()).device

    def predict(
        self,
        past_target: List[np.ndarray],
        feat_dynamic_real=None,
        past_feat_dynamic_real=None,
    ) -> np.ndarray:

        data_entry = {
            "past_target": past_target,
            "feat_dynamic_real": feat_dynamic_real,
            "past_feat_dynamic_real": past_feat_dynamic_real,
        }

        data_entry["past_observed_target"] = [~np.isnan(x) for x in past_target]
        if feat_dynamic_real:
            data_entry["observed_feat_dynamic_real"] = [
                ~np.isnan(x) for x in feat_dynamic_real
            ]
        else:
            data_entry["observed_feat_dynamic_real"] = None

        if past_feat_dynamic_real:
            data_entry["past_observed_feat_dynamic_real"] = [
                ~np.isnan(x) for x in past_feat_dynamic_real
            ]
        else:
            data_entry["past_observed_feat_dynamic_real"] = None

        impute = CausalMeanImputation()

        def process_sample(sample):
            arr = np.asarray(sample)
            if arr.ndim == 1:
                arr = arr[:, np.newaxis]
            if np.issubdtype(arr.dtype, np.number) and np.isnan(arr).any():
                arr = impute(arr)
            return arr

        for key, value in data_entry.items():
            if value is not None:
                data_entry[key] = [process_sample(sample) for sample in value]

        data_entry["past_is_pad"] = np.zeros(
            (len(data_entry["past_target"]), self.hparams.context_length), dtype=bool
        )

        for key in data_entry:
            if data_entry[key] is not None and isinstance(data_entry[key], list):
                for idx in range(len(data_entry[key])):
                    if data_entry[key][idx].shape[0] > self.hparams.context_length:
                        data_entry[key][idx] = data_entry[key][idx][
                            -self.hparams.context_length :, :
                        ]
                    else:
                        pad_length = (
                            self.hparams.context_length - data_entry[key][idx].shape[0]
                        )
                        pad_block = np.full(
                            (pad_length, 1),
                            data_entry[key][idx][0],
                            dtype=data_entry[key][idx].dtype,
                        )
                        data_entry[key][idx] = np.concatenate(
                            [pad_block, data_entry[key][idx]], axis=0
                        )
                        if key == "past_target":
                            data_entry["past_is_pad"][idx, :pad_length] = True

        for k in ["past_target", "feat_dynamic_real", "past_feat_dynamic_real"]:
            if data_entry[k] is not None:
                data_entry[k] = torch.tensor(
                    np.array(data_entry[k]), device=self.device, dtype=torch.float32
                )

        for k in [
            "past_observed_target",
            "observed_feat_dynamic_real",
            "past_observed_feat_dynamic_real",
            "past_is_pad",
        ]:
            if data_entry[k] is not None:
                data_entry[k] = torch.tensor(
                    np.array(data_entry[k]), device=self.device, dtype=torch.bool
                )

        with torch.no_grad():
            predictions = self(**data_entry).detach().cpu().numpy()
        return predictions

    @staticmethod
    def _patched_seq_pad(patch_size, x, dim, left=True, value=None):
        if dim >= 0:
            dim = -x.ndim + dim
        pad_length = -x.size(dim) % patch_size
        if left:
            pad = (pad_length, 0)
        else:
            pad = (0, pad_length)
        pad = (0, 0) * (abs(dim) - 1) + pad
        return torch.nn.functional.pad(x, pad, value=value)

    def _generate_time_id(self, patch_size, past_observed_target):
        past_seq_id = reduce(
            self._patched_seq_pad(patch_size, past_observed_target, -2, left=True),
            "... (seq patch) dim -> ... seq",
            "max",
            patch=patch_size,
        )
        past_seq_id = torch.clamp(
            past_seq_id.cummax(dim=-1).values.cumsum(dim=-1) - 1, min=0
        )
        batch_shape = " ".join(map(str, past_observed_target.shape[:-2]))
        future_seq_id = (
            repeat(
                torch.arange(
                    self.prediction_token_length(patch_size),
                    device=past_observed_target.device,
                ),
                f"prediction -> {batch_shape} prediction",
            )
            + past_seq_id.max(dim=-1, keepdim=True).values
            + 1
        )
        return past_seq_id, future_seq_id

    def _convert(
        self,
        patch_size,
        past_target,
        past_observed_target,
        past_is_pad,
        future_target=None,
        future_observed_target=None,
        future_is_pad=None,
        feat_dynamic_real=None,
        observed_feat_dynamic_real=None,
        past_feat_dynamic_real=None,
        past_observed_feat_dynamic_real=None,
    ):
        batch_shape = past_target.shape[:-2]
        device = past_target.device

        target = []
        observed_mask = []
        sample_id = []
        time_id = []
        variate_id = []
        prediction_mask = []
        dim_count = 0

        past_seq_id, future_seq_id = self._generate_time_id(
            patch_size, past_observed_target
        )

        if future_target is None:
            future_target = torch.zeros(
                batch_shape
                + (
                    self.hparams.prediction_length,
                    past_target.shape[-1],
                ),
                dtype=past_target.dtype,
                device=device,
            )
        target.extend(
            [
                torch.nn.functional.pad(
                    rearrange(
                        self._patched_seq_pad(patch_size, past_target, -2, left=True),
                        "... (seq patch) dim -> ... (dim seq) patch",
                        patch=patch_size,
                    ),
                    (0, 0),
                ),
                torch.nn.functional.pad(
                    rearrange(
                        self._patched_seq_pad(
                            patch_size, future_target, -2, left=False
                        ),
                        "... (seq patch) dim -> ... (dim seq) patch",
                        patch=patch_size,
                    ),
                    (0, 0),
                ),
            ]
        )
        if future_observed_target is None:
            future_observed_target = torch.ones(
                batch_shape
                + (
                    self.hparams.prediction_length,
                    past_observed_target.shape[-1],
                ),
                dtype=torch.bool,
                device=device,
            )
        observed_mask.extend(
            [
                torch.nn.functional.pad(
                    rearrange(
                        self._patched_seq_pad(
                            patch_size, past_observed_target, -2, left=True
                        ),
                        "... (seq patch) dim -> ... (dim seq) patch",
                        patch=patch_size,
                    ),
                    (0, 0),
                ),
                torch.nn.functional.pad(
                    rearrange(
                        self._patched_seq_pad(
                            patch_size, future_observed_target, -2, left=False
                        ),
                        "... (seq patch) dim -> ... (dim seq) patch",
                        patch=patch_size,
                    ),
                    (0, 0),
                ),
            ]
        )
        if future_is_pad is None:
            future_is_pad = torch.zeros(
                batch_shape + (self.hparams.prediction_length,),
                dtype=torch.long,
                device=device,
            )
        sample_id.extend(
            [
                repeat(
                    reduce(
                        (
                            self._patched_seq_pad(
                                patch_size, past_is_pad, -1, left=True, value=1
                            )
                            == 0
                        ).int(),
                        "... (seq patch) -> ... seq",
                        "max",
                        patch=patch_size,
                    ),
                    "... seq -> ... (dim seq)",
                    dim=past_target.shape[-1],
                ),
                repeat(
                    reduce(
                        (
                            self._patched_seq_pad(
                                patch_size, future_is_pad, -1, left=False, value=1
                            )
                            == 0
                        ).int(),
                        "... (seq patch) -> ... seq",
                        "max",
                        patch=patch_size,
                    ),
                    "... seq -> ... (dim seq)",
                    dim=past_target.shape[-1],
                ),
            ]
        )
        time_id.extend(
            [past_seq_id] * past_target.shape[-1]
            + [future_seq_id] * past_target.shape[-1]
        )
        variate_id.extend(
            [
                repeat(
                    torch.arange(past_target.shape[-1], device=device) + dim_count,
                    f"dim -> {' '.join(map(str, batch_shape))} (dim past)",
                    past=self.context_token_length(patch_size),
                ),
                repeat(
                    torch.arange(past_target.shape[-1], device=device) + dim_count,
                    f"dim -> {' '.join(map(str, batch_shape))} (dim future)",
                    future=self.prediction_token_length(patch_size),
                ),
            ]
        )
        dim_count += past_target.shape[-1]
        prediction_mask.extend(
            [
                torch.zeros(
                    batch_shape
                    + (self.context_token_length(patch_size) * past_target.shape[-1],),
                    dtype=torch.bool,
                    device=device,
                ),
                torch.ones(
                    batch_shape
                    + (
                        self.prediction_token_length(patch_size)
                        * past_target.shape[-1],
                    ),
                    dtype=torch.bool,
                    device=device,
                ),
            ]
        )

        if feat_dynamic_real is not None:
            if observed_feat_dynamic_real is None:
                raise ValueError(
                    "observed_feat_dynamic_real must be provided if feat_dynamic_real is provided"
                )

            target.extend(
                [
                    torch.nn.functional.pad(
                        rearrange(
                            self._patched_seq_pad(
                                patch_size,
                                feat_dynamic_real[
                                    ..., : self.hparams.context_length, :
                                ],
                                -2,
                                left=True,
                            ),
                            "... (seq patch) dim -> ... (dim seq) patch",
                            patch=patch_size,
                        ),
                        (0, 0),
                    ),
                    torch.nn.functional.pad(
                        rearrange(
                            self._patched_seq_pad(
                                patch_size,
                                feat_dynamic_real[
                                    ..., self.hparams.context_length :, :
                                ],
                                -2,
                                left=False,
                            ),
                            "... (seq patch) dim -> ... (dim seq) patch",
                            patch=patch_size,
                        ),
                        (0, 0),
                    ),
                ]
            )
            observed_mask.extend(
                [
                    torch.nn.functional.pad(
                        rearrange(
                            self._patched_seq_pad(
                                patch_size,
                                observed_feat_dynamic_real[
                                    ..., : self.hparams.context_length, :
                                ],
                                -2,
                                left=True,
                            ),
                            "... (seq patch) dim -> ... (dim seq) patch",
                            patch=patch_size,
                        ),
                        (0, 0),
                    ),
                    torch.nn.functional.pad(
                        rearrange(
                            self._patched_seq_pad(
                                patch_size,
                                observed_feat_dynamic_real[
                                    ..., self.hparams.context_length :, :
                                ],
                                -2,
                                left=False,
                            ),
                            "... (seq patch) dim -> ... (dim seq) patch",
                            patch=patch_size,
                        ),
                        (0, 0),
                    ),
                ]
            )
            sample_id.extend(
                [
                    repeat(
                        reduce(
                            (
                                self._patched_seq_pad(
                                    patch_size, past_is_pad, -1, left=True
                                )
                                == 0
                            ).int(),
                            "... (seq patch) -> ... seq",
                            "max",
                            patch=patch_size,
                        ),
                        "... seq -> ... (dim seq)",
                        dim=feat_dynamic_real.shape[-1],
                    ),
                    torch.ones(
                        batch_shape
                        + (
                            self.prediction_token_length(patch_size)
                            * feat_dynamic_real.shape[-1],
                        ),
                        dtype=torch.long,
                        device=device,
                    ),
                ]
            )
            time_id.extend(
                [past_seq_id] * feat_dynamic_real.shape[-1]
                + [future_seq_id] * feat_dynamic_real.shape[-1]
            )
            variate_id.extend(
                [
                    repeat(
                        torch.arange(feat_dynamic_real.shape[-1], device=device)
                        + dim_count,
                        f"dim -> {' '.join(map(str, batch_shape))} (dim past)",
                        past=self.context_token_length(patch_size),
                    ),
                    repeat(
                        torch.arange(feat_dynamic_real.shape[-1], device=device)
                        + dim_count,
                        f"dim -> {' '.join(map(str, batch_shape))} (dim future)",
                        future=self.prediction_token_length(patch_size),
                    ),
                ]
            )
            dim_count += feat_dynamic_real.shape[-1]
            prediction_mask.extend(
                [
                    torch.zeros(
                        batch_shape
                        + (
                            self.context_token_length(patch_size)
                            * feat_dynamic_real.shape[-1],
                        ),
                        dtype=torch.bool,
                        device=device,
                    ),
                    torch.zeros(
                        batch_shape
                        + (
                            self.prediction_token_length(patch_size)
                            * feat_dynamic_real.shape[-1],
                        ),
                        dtype=torch.bool,
                        device=device,
                    ),
                ]
            )

        if past_feat_dynamic_real is not None:
            if past_observed_feat_dynamic_real is None:
                raise ValueError(
                    "past_observed_feat_dynamic_real must be provided if past_feat_dynamic_real is provided"
                )
            target.append(
                torch.nn.functional.pad(
                    rearrange(
                        self._patched_seq_pad(
                            patch_size, past_feat_dynamic_real, -2, left=True
                        ),
                        "... (seq patch) dim -> ... (dim seq) patch",
                        patch=patch_size,
                    ),
                    (0, 0),
                )
            )
            observed_mask.append(
                torch.nn.functional.pad(
                    rearrange(
                        self._patched_seq_pad(
                            patch_size, past_observed_feat_dynamic_real, -2, left=True
                        ),
                        "... (seq patch) dim -> ... (dim seq) patch",
                        patch=patch_size,
                    ),
                    (0, 0),
                )
            )
            sample_id.append(
                repeat(
                    reduce(
                        (
                            self._patched_seq_pad(
                                patch_size, past_is_pad, -1, left=True
                            )
                            == 0
                        ).int(),
                        "... (seq patch) -> ... seq",
                        "max",
                        patch=patch_size,
                    ),
                    "... seq -> ... (dim seq)",
                    dim=past_feat_dynamic_real.shape[-1],
                )
            )
            time_id.extend([past_seq_id] * past_feat_dynamic_real.shape[-1])

            variate_id.append(
                repeat(
                    torch.arange(past_feat_dynamic_real.shape[-1], device=device)
                    + dim_count,
                    f"dim -> {' '.join(map(str, batch_shape))} (dim past)",
                    past=self.context_token_length(patch_size),
                )
            )
            dim_count += past_feat_dynamic_real.shape[-1]
            prediction_mask.append(
                torch.zeros(
                    batch_shape
                    + (
                        self.context_token_length(patch_size)
                        * past_feat_dynamic_real.shape[-1],
                    ),
                    dtype=torch.bool,
                    device=device,
                )
            )

        target = torch.cat(target, dim=-2)
        observed_mask = torch.cat(observed_mask, dim=-2)
        sample_id = torch.cat(sample_id, dim=-1)
        time_id = torch.cat(time_id, dim=-1)
        variate_id = torch.cat(variate_id, dim=-1)
        prediction_mask = torch.cat(prediction_mask, dim=-1)
        return (
            target,
            observed_mask,
            sample_id,
            time_id,
            variate_id,
            prediction_mask,
        )

    def _format_preds(self, num_quantiles, patch_size, preds, target_dim):
        start = target_dim * self.context_token_length(patch_size)
        end = start + target_dim * self.prediction_token_length(patch_size)
        preds = preds[..., start:end, :num_quantiles, :patch_size]
        preds = rearrange(
            preds,
            "... (dim seq) num_quantiles patch -> ... num_quantiles (seq patch) dim",
            dim=target_dim,
        )[..., : self.hparams.prediction_length, :]
        return preds.squeeze(-1)



#  Copyright (c) 2024, Salesforce, Inc.
#  SPDX-License-Identifier: Apache-2.0

from typing import Optional

import torch
from torch import nn


class RMSNorm(nn.Module):
    def __init__(
        self,
        normalized_shape: int | list[int] | torch.Size,
        eps: float = 1e-5,
        weight: bool = True,
        dtype: Optional[torch.dtype] = None,
    ):
        super().__init__()
        if isinstance(normalized_shape, int):
            normalized_shape = (normalized_shape,)

        self.normalized_shape = normalized_shape
        self.eps = eps
        self.mean_dim = tuple(range(-len(normalized_shape), 0))

        if weight:
            self.weight = torch.nn.Parameter(torch.ones(normalized_shape, dtype=dtype))
        else:
            self.register_parameter("weight", None)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        output = x * torch.rsqrt(
            x.pow(2).mean(dim=self.mean_dim, keepdim=True) + self.eps
        )
        if self.weight is not None:
            return output * self.weight
        return output

    def extra_repr(self) -> str:
        return (
            f"normalized_shape={self.normalized_shape}, "
            f"eps={self.eps}, "
            f"weight={self.weight is not None}"
        )



#  Copyright (c) 2024, Salesforce, Inc.
#  SPDX-License-Identifier: Apache-2.0

import math
from collections.abc import Callable
from functools import partial
from typing import Optional

import torch
import torch.nn.functional as F
from einops import rearrange, repeat
from torch import nn

from .position import AttentionBias, QueryKeyProjection


class GroupedQueryAttention(nn.Module):
    def __init__(
        self,
        dim: int,
        num_heads: int,
        num_groups: int,
        bias: bool = True,
        norm_layer: Optional[type[nn.Module] | partial[nn.Module]] = nn.LayerNorm,
        softmax_scale: Optional[float] = None,
        attn_dropout_p: float = 0.0,
        var_attn_bias: Optional[Callable[[], AttentionBias]] = None,
        time_attn_bias: Optional[Callable[[], AttentionBias]] = None,
        var_qk_proj: Optional[Callable[[], QueryKeyProjection]] = None,
        time_qk_proj: Optional[Callable[[], QueryKeyProjection]] = None,
    ):
        super().__init__()
        assert num_heads > 0 and dim % num_heads == 0
        assert (num_heads % num_groups == 0) and (num_heads >= num_groups)

        self.num_heads = num_heads
        self.num_groups = num_groups
        self.head_dim = dim // num_heads
        self.heads_per_group = num_heads // num_groups
        self.var_attn_bias = var_attn_bias() if var_attn_bias is not None else None
        self.time_attn_bias = time_attn_bias() if time_attn_bias is not None else None
        self.var_qk_proj = var_qk_proj() if var_qk_proj is not None else None
        self.time_qk_proj = time_qk_proj() if time_qk_proj is not None else None

        self.softmax_scale = softmax_scale or 1 / math.sqrt(self.head_dim)

        self.q_proj = nn.Linear(dim, dim, bias=bias)
        self.k_proj = nn.Linear(dim, self.head_dim * num_groups, bias=bias)
        self.v_proj = nn.Linear(dim, self.head_dim * num_groups, bias=bias)
        self.q_norm = (
            norm_layer(self.head_dim) if norm_layer is not None else nn.Identity()
        )
        self.k_norm = (
            norm_layer(self.head_dim) if norm_layer is not None else nn.Identity()
        )
        self.attn_dropout_p = attn_dropout_p
        self.out_proj = nn.Linear(dim, dim, bias=bias)

    def _get_var_id(self, query, key, query_var_id, kv_var_id):
        if self.var_attn_bias is not None or self.var_qk_proj is not None:
            if query_var_id is None:
                query_var_id = repeat(
                    torch.zeros((), device=query.device, dtype=torch.long),
                    f" -> {' '.join(map(str, query.shape[:-4]))} 1 1 {query.shape[-2]}",
                )
            else:
                query_var_id = rearrange(query_var_id, "... q_len -> ... 1 1 q_len")

            if kv_var_id is None:
                kv_var_id = repeat(
                    torch.zeros((), device=key.device, dtype=torch.long),
                    f" -> {' '.join(map(str, key.shape[:-4]))} 1 1 {key.shape[-2]}",
                )
            else:
                kv_var_id = rearrange(kv_var_id, "... kv_len -> ... 1 1 kv_len")

        return query_var_id, kv_var_id

    def _get_time_id(self, query, key, query_time_id, kv_time_id):
        if self.time_attn_bias is not None or self.time_qk_proj is not None:
            if query_time_id is None:
                query_time_id = repeat(
                    torch.arange(
                        query.shape[-2], device=query.device, dtype=torch.long
                    ),
                    f"q_len -> {' '.join(map(str, query.shape[:-4]))} 1 1 q_len",
                )
            else:
                query_time_id = rearrange(query_time_id, "... q_len -> ... 1 1 q_len")

            if kv_time_id is None:
                kv_time_id = repeat(
                    torch.arange(key.shape[-2], device=key.device, dtype=torch.long),
                    f"kv_len -> {' '.join(map(str, key.shape[:-4]))} 1 1 kv_len",
                )
            else:
                kv_time_id = rearrange(kv_time_id, "... kv_len-> ... 1 1 kv_len")

        return query_time_id, kv_time_id

    def _update_attn_mask(
        self, attn_mask, query, key,
        query_var_id=None, kv_var_id=None,
        query_time_id=None, kv_time_id=None,
    ):
        if attn_mask is not None:
            attn_mask = rearrange(
                attn_mask,
                "... q_len kv_len -> ... 1 1 q_len kv_len",
            )

        attn_bias = 0
        if self.var_attn_bias is not None:
            attn_bias = attn_bias + self.var_attn_bias(
                query, key, query_id=query_var_id, kv_id=kv_var_id,
            )

        if self.time_attn_bias is not None:
            attn_bias = attn_bias + self.time_attn_bias(
                query, key, query_id=query_time_id, kv_id=kv_time_id,
            )

        attn_mask = (
            attn_mask
            if isinstance(attn_bias, int)
            else (
                attn_bias
                if attn_mask is None
                else attn_bias.masked_fill(attn_mask.logical_not(), float("-inf"))
            )
        )
        return attn_mask

    def _qk_proj(
        self, query, key,
        query_var_id=None, kv_var_id=None,
        query_time_id=None, kv_time_id=None,
    ):
        if self.var_qk_proj is not None:
            query, key = self.var_qk_proj(
                query, key, query_id=query_var_id, kv_id=kv_var_id
            )

        if self.time_qk_proj is not None:
            query, key = self.time_qk_proj(
                query, key, query_id=query_time_id, kv_id=kv_time_id
            )

        return query, key

    def forward(
        self, query, key, value,
        attn_mask=None,
        query_var_id=None, kv_var_id=None,
        query_time_id=None, kv_time_id=None,
    ):
        query = self.q_proj(query)
        key = self.k_proj(key)
        value = self.v_proj(value)

        query = self.q_norm(
            rearrange(
                query,
                "... q_len (group hpg dim) -> ... group hpg q_len dim",
                group=self.num_groups,
                hpg=self.heads_per_group,
            )
        )
        key = self.k_norm(
            repeat(
                key,
                "... kv_len (group dim) -> ... group hpg kv_len dim",
                group=self.num_groups,
                hpg=self.heads_per_group,
            )
        )
        value = repeat(
            value,
            "... kv_len (group dim) -> ... group hpg kv_len dim",
            group=self.num_groups,
            hpg=self.heads_per_group,
        )

        query_var_id, kv_var_id = self._get_var_id(query, key, query_var_id, kv_var_id)
        query_time_id, kv_time_id = self._get_time_id(
            query, key, query_time_id, kv_time_id,
        )

        attn_mask = self._update_attn_mask(
            attn_mask, query, key,
            query_var_id=query_var_id, kv_var_id=kv_var_id,
            query_time_id=query_time_id, kv_time_id=kv_time_id,
        )

        query, key = self._qk_proj(
            query, key,
            query_var_id=query_var_id, kv_var_id=kv_var_id,
            query_time_id=query_time_id, kv_time_id=kv_time_id,
        )

        out = F.scaled_dot_product_attention(
            query, key, value,
            attn_mask=attn_mask,
            dropout_p=self.attn_dropout_p,
            scale=self.softmax_scale,
        )
        out = rearrange(out, "... group hpg q_len dim -> ... q_len (group hpg dim)")
        return self.out_proj(out)



"""
moirai_sklearn - A scikit-learn-like interface for Moirai time series forecasting.

Uses native quantile outputs to compute mean, median, mode, std, etc.

Usage:
    from moirai_sklearn import MoiraiForecaster
    
    model = MoiraiForecaster()
    predictions = model.predict(my_time_series, horizon=30)
"""

from .forecaster import MoiraiForecaster

__version__ = "0.2.0"
__all__ = ["MoiraiForecaster"]



#  Copyright (c) 2024, Salesforce, Inc.
#  SPDX-License-Identifier: Apache-2.0

import abc
import math
from functools import cached_property
from typing import Any, Optional

import torch
from einops import einsum, rearrange, repeat
from torch import nn


# --- Attention Bias ---

class AttentionBias(nn.Module, abc.ABC):
    def __init__(self, dim: int, num_heads: int, num_groups: int):
        super().__init__()
        assert num_heads > 0 and dim % num_heads == 0
        assert (num_heads % num_groups == 0) and (num_heads >= num_groups)

        self.num_heads = num_heads
        self.num_groups = num_groups
        self.heads_per_group = num_heads // num_groups
        self.head_dim = dim // num_heads

    @abc.abstractmethod
    def forward(self, query, key, query_id, kv_id): ...


class BinaryAttentionBias(AttentionBias):
    def __init__(self, dim: int, num_heads: int, num_groups: int):
        super().__init__(dim, num_heads, num_groups)
        self.emb = nn.Embedding(num_embeddings=2, embedding_dim=self.num_heads)

    def forward(self, query, key, query_id, kv_id):
        ind = torch.eq(query_id.unsqueeze(-1), kv_id.unsqueeze(-2))
        weight = rearrange(self.emb.weight, "two num_heads -> two num_heads 1 1")
        bias = rearrange(
            ~ind * weight[:1] + ind * weight[1:],
            "... 1 (group hpg) q_len kv_len -> ... group hpg q_len kv_len",
            group=self.num_groups,
            hpg=self.heads_per_group,
        )
        return bias


# --- Projections ---

class Projection(nn.Module, abc.ABC):
    def __init__(self, proj_width: int, num_heads: int, num_groups: int, **kwargs: Any):
        super().__init__()
        self.proj_width = proj_width
        self.num_heads = num_heads
        self.num_groups = num_groups
        self.heads_per_group = num_heads // num_groups

    @abc.abstractmethod
    def forward(self, x, seq_id=None): ...


class RotaryProjection(Projection):
    def __init__(
        self,
        *,
        proj_width: int,
        num_heads: int,
        num_groups: int,
        max_len: int = 512,
        base: int = 10000,
    ):
        super().__init__(proj_width, num_heads, num_groups)
        assert self.proj_width % 2 == 0
        self.register_buffer(
            "theta",
            1.0
            / torch.pow(
                base,
                torch.arange(0, self.proj_width, 2, dtype=torch.float)
                / self.proj_width,
            ),
            persistent=False,
        )
        self.register_buffer("cos", None, persistent=False)
        self.register_buffer("sin", None, persistent=False)
        self._init_freq(max_len=max_len)

    def _init_freq(self, max_len: int):
        if self.cos is None or self.cos.size(-2) < max_len:
            position = torch.arange(
                max_len, device=self.theta.device, dtype=self.theta.dtype
            )
            m_theta = einsum(position, self.theta, "length, width -> length width")
            m_theta = repeat(m_theta, "length width -> length (width 2)")
            self.register_buffer("cos", torch.cos(m_theta), persistent=False)
            self.register_buffer("sin", torch.sin(m_theta), persistent=False)

    @staticmethod
    def _rotate(x):
        x1, x2 = rearrange(x, "... (dim r) -> r ... dim", r=2)
        return rearrange([-x2, x1], "r ... dim -> ... (dim r)", r=2)  # noqa

    def forward(self, x, seq_id=None):
        self._init_freq(max_len=seq_id.max() + 1)
        rot_cos = self.cos[seq_id]
        rot_sin = self.sin[seq_id]
        return rot_cos * x + rot_sin * self._rotate(x)


class QueryKeyProjection(nn.Module):
    def __init__(
        self,
        dim: int,
        num_heads: int,
        num_groups: int,
        proj_layer: type[Projection],
        kwargs: Optional[dict[str, Any]] = None,
        key_proj_layer: Optional[type[Projection]] = None,
        key_kwargs: Optional[dict[str, Any]] = None,
        partial_factor: Optional[tuple[float, float]] = None,
    ):
        super().__init__()
        if partial_factor is not None:
            assert 0.0 <= partial_factor[0] < partial_factor[1] <= 1.0
        assert num_heads > 0 and dim % num_heads == 0
        assert (num_heads % num_groups == 0) and (num_heads >= num_groups)

        self.head_dim = dim // num_heads
        self.partial_factor = partial_factor
        self.query_proj = proj_layer(
            proj_width=self.proj_width,
            num_heads=num_heads,
            num_groups=num_groups,
            **(kwargs or {}),
        )
        if key_proj_layer is None:
            self.key_proj = self.query_proj
        else:
            self.key_proj = key_proj_layer(
                proj_width=self.proj_width,
                num_heads=num_heads,
                num_groups=num_groups,
                **(key_kwargs or {}),
            )

    @cached_property
    def proj_width(self) -> int:
        if self.partial_factor is None:
            return self.head_dim
        return int(self.head_dim * (self.partial_factor[1] - self.partial_factor[0]))

    @cached_property
    def split_sizes(self) -> tuple[int, int, int]:
        if self.partial_factor is None:
            return 0, self.head_dim, 0
        return (
            int(self.partial_factor[0] * self.head_dim),
            self.proj_width,
            int((1.0 - self.partial_factor[1]) * self.head_dim),
        )

    def forward(self, query, key, query_id=None, kv_id=None):
        if self.partial_factor is not None:
            queries = list(query.split(self.split_sizes, dim=-1))
            keys = list(key.split(self.split_sizes, dim=-1))
            queries[1] = self.query_proj(queries[1], seq_id=query_id)
            keys[1] = self.key_proj(keys[1], seq_id=kv_id)
            query = torch.cat(queries, dim=-1)
            key = torch.cat(keys, dim=-1)
        else:
            query = self.query_proj(query, seq_id=query_id)
            key = self.key_proj(key, seq_id=kv_id)
        return query, key



"""
Moirai Forecaster - A scikit-learn-like wrapper for Moirai time series models.

Uses the native quantile predictions from Moirai to provide:
- predict_mean: estimated from quantiles
- predict_median: direct from p50
- predict_mode: estimated from quantile density
- predict_quantile: interpolated from model quantiles
- predict_interval: prediction intervals from quantiles
- predict_std: estimated from IQR
"""

import numpy as np
import pandas as pd
from typing import Optional, Union, List


class MoiraiForecaster:
    """
    A simple scikit-learn-like interface for Moirai time series forecasting.
    
    Uses native quantile outputs from Moirai (p10, p20, ..., p90) to compute
    all distribution statistics directly.
    
    Parameters
    ----------
    model_name : str, default="Salesforce/moirai-2.0-R-small"
        Name of the pretrained Moirai model from HuggingFace.
        Options include:
        - "Salesforce/moirai-2.0-R-small"
        - "Salesforce/moirai-2.0-R-base"
        - "Salesforce/moirai-2.0-R-large"
        
    context_length : int, default=1000
        Number of past observations to use as context.
        
    device : str, default="auto"
        Device to run the model on ("auto", "cpu", "cuda").
    
    Examples
    --------
    >>> from moirai_sklearn import MoiraiForecaster
    >>> import numpy as np
    >>> 
    >>> ts = np.sin(np.linspace(0, 10, 100)) + np.random.randn(100) * 0.1
    >>> model = MoiraiForecaster()
    >>> predictions = model.predict(ts, horizon=30)
    """
    
    def __init__(
        self,
        model_name: str = "Salesforce/moirai-2.0-R-small",
        context_length: int = 1000,
        device: str = "auto",
    ):
        self.model_name = model_name
        self.context_length = context_length
        self.device = device
        
        self._model = None
        self._module = None
        self._quantile_levels = None
        
    def _ensure_model_loaded(self, horizon: int):
        """Lazy load the model on first use."""
        if self._model is None or self._model.hparams.prediction_length != horizon:
            from ._vendor import Moirai2Forecast, Moirai2Module
            
            if self._module is None:
                self._module = Moirai2Module.from_pretrained(self.model_name)
            
            self._model = Moirai2Forecast(
                module=self._module,
                prediction_length=horizon,
                context_length=self.context_length,
                target_dim=1,
                feat_dynamic_real_dim=0,
                past_feat_dynamic_real_dim=0,
            )
            
            if self.device != "auto":
                self._model = self._model.to(self.device)
                
            self._quantile_levels = np.array(self._module.quantile_levels)
    
    def _prepare_input(
        self, 
        data: Union[np.ndarray, pd.Series, pd.DataFrame],
    ) -> np.ndarray:
        """Convert input to numpy array."""
        if isinstance(data, pd.Series):
            arr = data.values
        elif isinstance(data, np.ndarray):
            arr = data.flatten()
        elif isinstance(data, pd.DataFrame):
            # Use first column
            arr = data.iloc[:, 0].values
        else:
            raise ValueError(f"Unsupported data type: {type(data)}")
        
        return arr.astype(np.float32)
    
    def _raw_predict(
        self, 
        data: Union[np.ndarray, pd.Series, pd.DataFrame],
        horizon: int,
    ) -> np.ndarray:
        """
        Get raw quantile predictions from the model.
        
        Returns
        -------
        predictions : np.ndarray of shape (num_quantiles, horizon)
            Quantile predictions at each future time step.
        """
        self._ensure_model_loaded(horizon)
        arr = self._prepare_input(data)
        
        # Model.predict takes a list of arrays
        # Returns shape: (batch=1, num_quantiles, horizon)
        preds = self._model.predict(past_target=[arr])
        
        # Return shape: (num_quantiles, horizon)
        return preds[0]
    
    def predict(
        self, 
        data: Union[np.ndarray, pd.Series, pd.DataFrame],
        horizon: int,
    ) -> np.ndarray:
        """
        Predict future values (returns median/p50).
        
        Parameters
        ----------
        data : array-like
            Historical time series data.
        horizon : int
            Number of steps to forecast.
            
        Returns
        -------
        predictions : np.ndarray of shape (horizon,)
            Point predictions (median).
        """
        return self.predict_median(data, horizon)
    
    def predict_median(
        self, 
        data: Union[np.ndarray, pd.Series, pd.DataFrame],
        horizon: int,
    ) -> np.ndarray:
        """
        Predict median (p50) directly from model quantiles.
        """
        preds = self._raw_predict(data, horizon)
        median_idx = np.argmin(np.abs(self._quantile_levels - 0.5))
        return preds[median_idx]
    
    def predict_mean(
        self, 
        data: Union[np.ndarray, pd.Series, pd.DataFrame],
        horizon: int,
    ) -> np.ndarray:
        """
        Estimate mean as average of all quantiles.
        
        Simple approximation: mean ≈ average of [p10, p20, ..., p90]
        """
        preds = self._raw_predict(data, horizon)
        # Take mean across all quantiles
        return np.mean(preds, axis=0)
    
    def predict_mode(
        self, 
        data: Union[np.ndarray, pd.Series, pd.DataFrame],
        horizon: int,
    ) -> np.ndarray:
        """
        Estimate mode as the median (p50).
        
        For symmetric distributions, mode ≈ median.
        """
        return self.predict_median(data, horizon)
    
    def predict_quantile(
        self, 
        data: Union[np.ndarray, pd.Series, pd.DataFrame],
        horizon: int,
        q: Union[float, List[float], np.ndarray] = 0.5,
    ) -> np.ndarray:
        """
        Predict quantile(s) from model outputs.
        
        Parameters
        ----------
        q : float or array-like
            Quantile(s) to get. Must be one of the model's quantile levels:
            [0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9]
            
        Returns
        -------
        quantiles : np.ndarray
            Shape (horizon,) if q is scalar, (horizon, n_quantiles) if q is array.
        """
        preds = self._raw_predict(data, horizon)
        q_arr = np.atleast_1d(q)
        
        # Find indices for requested quantiles
        results = np.zeros((horizon, len(q_arr)))
        for i, q_val in enumerate(q_arr):
            idx = np.argmin(np.abs(self._quantile_levels - q_val))
            if np.abs(self._quantile_levels[idx] - q_val) > 0.01:
                raise ValueError(
                    f"Quantile {q_val} not available. "
                    f"Available quantiles: {self._quantile_levels.tolist()}"
                )
            results[:, i] = preds[idx]
        
        if np.isscalar(q):
            return results[:, 0]
        return results
    
    def predict_interval(
        self, 
        data: Union[np.ndarray, pd.Series, pd.DataFrame],
        horizon: int,
        confidence: float = 0.8,
    ) -> np.ndarray:
        """
        Predict prediction interval from quantiles.
        
        Parameters
        ----------
        confidence : float, default=0.8
            Confidence level. Only 0.8 (p10-p90) and 0.6 (p20-p80) are available
            based on the model's quantile outputs.
        
        Returns
        -------
        intervals : np.ndarray of shape (horizon, 2)
            Lower and upper bounds [low, high].
        """
        if confidence == 0.8:
            return self.predict_quantile(data, horizon, q=[0.1, 0.9])
        elif confidence == 0.6:
            return self.predict_quantile(data, horizon, q=[0.2, 0.8])
        else:
            raise ValueError(
                f"Confidence level {confidence} not supported. "
                "Available: 0.8 (p10-p90) or 0.6 (p20-p80)"
            )
    
    def predict_std(
        self, 
        data: Union[np.ndarray, pd.Series, pd.DataFrame],
        horizon: int,
    ) -> np.ndarray:
        """
        Estimate standard deviation from p10-p90 range.
        
        For normal distribution: std ≈ (p90 - p10) / 2.56
        """
        preds = self._raw_predict(data, horizon)
        # p10 is at index 0, p90 is at index 8
        return (preds[8] - preds[0]) / 2.56
    
    def predict_all(
        self, 
        data: Union[np.ndarray, pd.Series, pd.DataFrame],
        horizon: int,
    ) -> pd.DataFrame:
        """
        Get all predictions in a DataFrame.
        
        Returns
        -------
        df : pd.DataFrame
            Columns: step, mean, median, std, p10, p20, ..., p90
        """
        preds = self._raw_predict(data, horizon)
        
        result = {
            'step': np.arange(1, horizon + 1),
            'mean': self.predict_mean(data, horizon),
            'median': self.predict_median(data, horizon),
            'std': self.predict_std(data, horizon),
        }
        
        # Add all model quantiles
        for i, q in enumerate(self._quantile_levels):
            result[f'p{int(q*100)}'] = preds[i]
        
        return pd.DataFrame(result)



#  Copyright (c) 2024, Salesforce, Inc.
#  SPDX-License-Identifier: Apache-2.0

from typing import Optional

import torch


def packed_attention_mask(
    sample_id: torch.Tensor,
) -> torch.Tensor:
    sample_id = sample_id.unsqueeze(-1)
    attention_mask = sample_id.eq(sample_id.mT)
    return attention_mask


def packed_causal_attention_mask(
    sample_id: torch.Tensor,
    time_id: torch.Tensor,
) -> torch.Tensor:
    attention_mask = packed_attention_mask(sample_id)
    expanded_id1 = time_id.unsqueeze(-2)
    expanded_id2 = time_id.unsqueeze(-1)
    compare_res = expanded_id1 <= expanded_id2
    attention_mask = attention_mask * compare_res
    return attention_mask


def safe_div(
    numer: torch.Tensor,
    denom: torch.Tensor,
) -> torch.Tensor:
    return numer / torch.where(
        denom == 0,
        1.0,
        denom,
    )


def size_to_mask(
    max_size: int,
    sizes: torch.Tensor,
) -> torch.Tensor:
    mask = torch.arange(max_size, device=sizes.device)
    return torch.lt(mask, sizes.unsqueeze(-1))



#  Copyright (c) 2024, Salesforce, Inc.
#  SPDX-License-Identifier: Apache-2.0

from functools import partial

import torch
import torch.nn.functional as F
from huggingface_hub import PyTorchModelHubMixin
from torch import nn

from .torch_util import packed_causal_attention_mask
from .norm import RMSNorm
from .packed_scaler import PackedNOPScaler, PackedStdScaler
from .position import BinaryAttentionBias, QueryKeyProjection, RotaryProjection
from .transformer import TransformerEncoder
from .ts_embed import ResidualBlock


class Moirai2Module(
    nn.Module,
    PyTorchModelHubMixin,
):
    def __init__(
        self,
        d_model: int,
        d_ff: int,
        num_layers: int,
        patch_size: int,
        max_seq_len: int,
        attn_dropout_p: float,
        dropout_p: float,
        scaling: bool = True,
        num_predict_token: int = 1,
        quantile_levels: tuple[float] = (0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9),
    ):
        super().__init__()
        self.d_model = d_model
        self.num_layers = num_layers
        self.patch_size = patch_size
        self.num_predict_token = num_predict_token
        self.max_seq_len = max_seq_len
        self.scaling = scaling
        self.quantile_levels = quantile_levels
        self.num_quantiles = len(quantile_levels)

        self.scaler = PackedStdScaler() if scaling else PackedNOPScaler()
        self.in_proj = ResidualBlock(
            input_dims=patch_size * 2,
            hidden_dims=d_model,
            output_dims=d_model,
        )
        self.encoder = TransformerEncoder(
            d_model,
            num_layers,
            num_heads=None,
            pre_norm=True,
            attn_dropout_p=attn_dropout_p,
            dropout_p=dropout_p,
            norm_layer=RMSNorm,
            activation=F.silu,
            use_glu=True,
            use_qk_norm=True,
            var_attn_bias_layer=partial(BinaryAttentionBias),
            time_qk_proj_layer=partial(
                QueryKeyProjection,
                proj_layer=RotaryProjection,
                kwargs=dict(max_len=max_seq_len),
                partial_factor=(0.0, 0.5),
            ),
            shared_var_attn_bias=False,
            shared_time_qk_proj=True,
            d_ff=d_ff,
        )
        self.out_proj = ResidualBlock(
            input_dims=d_model,
            hidden_dims=d_model,
            output_dims=num_predict_token * self.num_quantiles * patch_size,
        )

    def forward(
        self,
        target,
        observed_mask,
        sample_id,
        time_id,
        variate_id,
        prediction_mask,
        training_mode=True,
    ):
        loc, scale = self.scaler(
            target,
            observed_mask * ~prediction_mask.unsqueeze(-1),
            sample_id,
            variate_id,
        )
        scaled_target = (target - loc) / scale
        input_tokens = torch.cat(
            [scaled_target, observed_mask.to(torch.float32)], dim=-1
        )
        reprs = self.in_proj(input_tokens)

        reprs = self.encoder(
            reprs,
            packed_causal_attention_mask(sample_id, time_id),
            time_id=time_id,
            var_id=variate_id,
        )
        preds = self.out_proj(reprs)
        if training_mode:
            return preds, scaled_target
        else:
            return preds * scale + loc



#  Copyright (c) 2024, Salesforce, Inc.
#  SPDX-License-Identifier: Apache-2.0

from typing import Optional

import torch
from einops import reduce
from torch import nn

from .torch_util import safe_div


class PackedScaler(nn.Module):
    def forward(
        self,
        target: torch.Tensor,
        observed_mask: Optional[torch.Tensor] = None,
        sample_id: Optional[torch.Tensor] = None,
        variate_id: Optional[torch.Tensor] = None,
    ):
        if observed_mask is None:
            observed_mask = torch.ones_like(target, dtype=torch.bool)
        if sample_id is None:
            sample_id = torch.zeros(
                target.shape[:-1], dtype=torch.long, device=target.device
            )
        if variate_id is None:
            variate_id = torch.zeros(
                target.shape[:-1], dtype=torch.long, device=target.device
            )

        loc, scale = self._get_loc_scale(
            target.double(), observed_mask, sample_id, variate_id
        )
        return loc.float(), scale.float()

    def _get_loc_scale(self, target, observed_mask, sample_id, variate_id):
        raise NotImplementedError


class PackedNOPScaler(PackedScaler):
    def _get_loc_scale(self, target, observed_mask, sample_id, variate_id):
        loc = torch.zeros_like(target, dtype=target.dtype)
        scale = torch.ones_like(target, dtype=target.dtype)
        return loc, scale


class PackedStdScaler(PackedScaler):
    def __init__(self, correction: int = 1, minimum_scale: float = 1e-5):
        super().__init__()
        self.correction = correction
        self.minimum_scale = minimum_scale

    def _get_loc_scale(self, target, observed_mask, sample_id, variate_id):
        id_mask = torch.logical_and(
            torch.eq(sample_id.unsqueeze(-1), sample_id.unsqueeze(-2)),
            torch.eq(variate_id.unsqueeze(-1), variate_id.unsqueeze(-2)),
        )
        tobs = reduce(
            id_mask * reduce(observed_mask, "... seq dim -> ... 1 seq", "sum"),
            "... seq1 seq2 -> ... seq1 1",
            "sum",
        )
        loc = reduce(
            id_mask * reduce(target * observed_mask, "... seq dim -> ... 1 seq", "sum"),
            "... seq1 seq2 -> ... seq1 1",
            "sum",
        )
        loc = safe_div(loc, tobs)
        var = reduce(
            id_mask
            * reduce(
                ((target - loc) ** 2) * observed_mask,
                "... seq dim -> ... 1 seq",
                "sum",
            ),
            "... seq1 seq2 -> ... seq1 1",
            "sum",
        )
        var = safe_div(var, (tobs - self.correction))
        scale = torch.sqrt(var + self.minimum_scale)
        loc[sample_id == 0] = 0
        scale[sample_id == 0] = 1
        return loc, scale



#  Copyright (c) 2024, Salesforce, Inc.
#  SPDX-License-Identifier: Apache-2.0

import math

import torch
from torch import nn


class ResidualBlock(nn.Module):
    def __init__(self, input_dims, hidden_dims, output_dims):
        super(ResidualBlock, self).__init__()
        self.input_dims = input_dims
        self.hidden_dims = hidden_dims
        self.output_dims = output_dims

        self.hidden_layer = nn.Linear(input_dims, hidden_dims)
        self.silu = nn.SiLU()
        self.output_layer = nn.Linear(hidden_dims, output_dims)
        self.residual_layer = nn.Linear(input_dims, output_dims)

        self.reset_parameters()

    def reset_parameters(self):
        for m in self.children():
            if isinstance(m, nn.Linear):
                nn.init.kaiming_uniform_(m.weight, a=math.sqrt(5))
                if m.bias is not None:
                    fan_in, _ = nn.init._calculate_fan_in_and_fan_out(m.weight)
                    bound = 1 / math.sqrt(fan_in) if fan_in > 0 else 0
                    nn.init.uniform_(m.bias, -bound, bound)

    def forward(self, x):
        hidden = self.hidden_layer(x)
        hidden = self.silu(hidden)
        output = self.output_layer(hidden)
        residual = self.residual_layer(x)
        return output + residual



#  Copyright (c) 2024, Salesforce, Inc.
#  SPDX-License-Identifier: Apache-2.0

from collections.abc import Callable
from typing import Optional

import torch
import torch.nn.functional as F
from torch import nn


class FeedForward(nn.Module):
    def __init__(
        self,
        in_dim: int,
        hidden_dim: Optional[int] = None,
        out_dim: Optional[int] = None,
        activation: Callable[[torch.Tensor], torch.Tensor] = F.gelu,
        bias: bool = True,
        ffn_dropout_p: float = 0.0,
    ):
        super().__init__()
        hidden_dim = hidden_dim or 4 * in_dim
        out_dim = out_dim or in_dim

        self.in_dim = in_dim
        self.hidden_dim = hidden_dim
        self.out_dim = out_dim
        self.bias = bias
        self.ffn_dropout_p = ffn_dropout_p

        self.fc1 = nn.Linear(in_dim, hidden_dim, bias=bias)
        self.fc2 = nn.Linear(hidden_dim, out_dim, bias=bias)
        self.dropout1 = nn.Dropout(ffn_dropout_p)
        self.dropout2 = nn.Dropout(ffn_dropout_p)
        self.activation = activation

    def forward(self, x, centroid=None):
        x = self._in_proj(x)
        return self.dropout2(self.fc2(self.dropout1(x)))

    def _in_proj(self, x):
        return self.activation(self.fc1(x))


class GatedLinearUnitFeedForward(FeedForward):
    def __init__(
        self,
        in_dim: int,
        hidden_dim: Optional[int] = None,
        out_dim: Optional[int] = None,
        activation: Callable[[torch.Tensor], torch.Tensor] = F.silu,
        bias: bool = True,
        ffn_dropout_p: float = 0.0,
    ):
        super().__init__(
            in_dim,
            hidden_dim=hidden_dim or self.adjust_hidden_dim(4 * in_dim),
            out_dim=out_dim,
            activation=activation,
            bias=bias,
            ffn_dropout_p=ffn_dropout_p,
        )
        self.fc_gate = nn.Linear(self.in_dim, self.hidden_dim, bias=self.bias)

    @staticmethod
    def adjust_hidden_dim(dim):
        return (int(dim * 2 / 3) + 7) // 8 * 8

    def _in_proj(self, x):
        return self.activation(self.fc_gate(x)) * self.fc1(x)


class MoEFeedForward(nn.Module):
    def __init__(
        self,
        num_experts: int,
        num_experts_per_token: int,
        in_dim: int,
        hidden_dim: Optional[int] = None,
        out_dim: Optional[int] = None,
        activation: Callable[[torch.Tensor], torch.Tensor] = F.silu,
        bias: bool = True,
        ffn_dropout_p: float = 0.0,
    ):
        super().__init__()
        self.num_experts = num_experts
        self.num_experts_per_token = num_experts_per_token

        self.experts = nn.ModuleList(
            [
                GatedLinearUnitFeedForward(
                    in_dim=in_dim,
                    hidden_dim=hidden_dim,
                    out_dim=out_dim,
                    activation=activation,
                    bias=bias,
                    ffn_dropout_p=ffn_dropout_p,
                )
                for _ in range(num_experts)
            ]
        )

    def forward(self, x, centroid=None):
        x_squashed = x.view(-1, x.shape[-1])

        centroid = centroid.to(x.device).type_as(x)
        if len(x.shape) > 3:
            x_temp = x.view(-1, x.shape[-2], x.shape[-1])
        else:
            x_temp = x
        centroid = centroid.unsqueeze(0).repeat(x_temp.shape[0], 1, 1)
        cdist = torch.cdist(x_temp, centroid)
        gate_logits = cdist.view(-1, cdist.shape[-1])

        weights, selected_experts = torch.topk(gate_logits, self.num_experts_per_token)
        weights = nn.functional.softmax(
            weights,
            dim=1,
            dtype=torch.float,
        ).type_as(x)

        results = torch.zeros_like(x_squashed)
        for i, expert in enumerate(self.experts):
            batch_idx, nth_expert = torch.where(selected_experts == i)
            results[batch_idx] += weights[batch_idx, nth_expert, None] * expert(
                x_squashed[batch_idx]
            )

        results = results.view_as(x)
        return results



#  Copyright (c) 2024, Salesforce, Inc.
#  SPDX-License-Identifier: Apache-2.0

import numpy as np


class LastValueImputation:
    def __init__(self, value=0.0):
        self.value = value

    def __call__(self, x):
        x = x.T
        x[0:1][np.isnan(x[0:1])] = self.value
        mask = np.isnan(x)
        idx = np.arange(len(x))
        if x.ndim == 2:
            idx = np.expand_dims(idx, axis=1)
        idx = np.where(~mask, idx, 0)
        idx = np.maximum.accumulate(idx, axis=0)
        if x.ndim == 2:
            x = x[idx, np.arange(x.shape[1])]
        else:
            x = x[idx]
        return x.T


class CausalMeanImputation:
    def __init__(self, value=0.0):
        self.value = value

    def __call__(self, x, value=0.0):
        mask = np.isnan(x).T

        last_value_imputation = LastValueImputation(self.value)
        x = last_value_imputation(x)
        mask[0] = False
        x = x.T

        if x.ndim == 1:
            adjusted_values_to_causality = np.concatenate((np.repeat(0.0, 1), x[:-1]))
            cumsum = np.cumsum(adjusted_values_to_causality)
            indices = np.linspace(0, len(x) - 1, len(x))
            indices[0] = 1
            ar_res = cumsum / indices
            x[mask] = ar_res[mask]
        else:
            adjusted_values_to_causality = np.vstack(
                (np.zeros((1, x.shape[1])), x[:-1, :])
            )
            cumsum = np.cumsum(adjusted_values_to_causality, axis=0)
            indices = np.linspace(0, len(x) - 1, len(x)).reshape(-1, 1)
            indices[0] = 1
            ar_res = cumsum / indices
            x[mask] = ar_res[mask]
        return x.T



#  Vendored from uni2ts (Salesforce, Inc.) - Apache 2.0 License
#  Stripped of heavyweight dependencies (lightning, gluonts, jax, jaxtyping, hydra, etc.)
#  Only requires: torch, einops, huggingface-hub, safetensors, numpy

from .moirai2_module import Moirai2Module
from .moirai2_forecast import Moirai2Forecast

__all__ = ["Moirai2Module", "Moirai2Forecast"]
